@@ -23,6 +23,9 @@ class UnsatisfiedAssumption(HypothesisException):
     If you're seeing this error something has gone wrong.
     """
 
+    def __init__(self, reason=None):
+        self.reason = reason
+
 
 class NoSuchExample(HypothesisException):
     """The condition we have been asked to satisfy appears to be always false.
@@ -129,6 +132,7 @@ def __getattr__(name):
             "instead, or `exceptiongroup.BaseExceptionGroup` before Python 3.11",
             since="2022-08-02",
             has_codemod=False,  # This would be a great PR though!
+            stacklevel=1,
         )
         return BaseExceptionGroup
 
@@ -168,3 +172,15 @@ class Found(Exception):
     """Signal that the example matches condition. Internal use only."""
 
     hypothesis_internal_never_escalate = True
+
+
+class RewindRecursive(Exception):
+    """Signal that the type inference should be rewound due to recursive types. Internal use only."""
+
+    def __init__(self, target):
+        self.target = target
+
+
+class SmallSearchSpaceWarning(HypothesisWarning):
+    """Indicates that an inferred strategy does not span the search space
+    in a meaningful way, for example by only creating default instances."""
